from collections import OrderedDict
import argparse
import os
import json
import re
import sys


class ArgParser(object):
    """
    Main class to parse arguments.
    Json parser for test parameters
    """
    def __init__(self, jsonfile):
        """
        Initial module. It will parse json file
        @param jsonfile: json file for test parameters
        """
        self._parser = argparse.ArgumentParser(conflict_handler='resolve', add_help=False)
        self._parser.add_argument('hosts', help='ESX hostname to test. Multiple hosts '
                                                'separated by space', nargs='+')
        # required group
        req_group_help = 'These arguments are needed. The list changed ' \
                         'depended on difference test cases.'
        self._req_group = self._parser.add_argument_group('Required arguments', req_group_help)
        self._req_group.add_argument('--test_ids', help='Test cases name defined in config.xml. '
                                     'Multiple test cases are separated by space',
                                     metavar='str', nargs='+', required=True)
        # optional group
        self._ops_group = self._parser.add_argument_group('Optional arguments',
                                                          'These params are optional')
        self._ops_group.add_argument('-h', '--help', help='Print this message and exit', action='help')
        # keep additional test param for special test cases.
        # key is keyword contain in test_ids, for example:
        # "Network", "stress", "IODM", "BOON", etc
        self._json_param = OrderedDict()
        self._parseTestParams(jsonfile)

    @property
    def args(self):
        """
        get the arguments from the command line in the dictionary form
        @return: dictionary of arguments
        """
        return self._args

    def _parseTestParams(self, jsonfile):
        """
        Parse params for single test cases, params are stored in json format.
        These params name are divided into two category.
        1. Name start with '--' is param flag, followed by string to describe
           what the flag is for. Example:
           "--test_ids=<str>": "Test name that from config.xml"
                              --> flag test_ids required input string
           "--num_vm=<int>": "Number of vm [default: 4]"
                            --> flag num_vm has default value 4 if not specify
           "--MTU=<int>": "MTU value [default: 1500] [choice: 1500,4000,6000]"
                          --> flag MTU has default value is 1500 if not specify
                          and option for this flag are 1500. 4000 and 6000.
                          Please make sure the default value is in one of
                          choice option
           "--skip_setup": "Skip Test Setup"
                          --> flag skip_setup default is False if not specify
                          otherwise is True

        2. Name does not start with '--' is the name for test case that has
           additional flags. The name is NOT case sensitive. Followed the name
           tag is the dictionary of additional flags. These flags follow the
           same rules as category 1. Example
           "network": {"--ipv6Server": "Server that need to test ipv6",
                       "--num_vm=<int>": "Number of vm [default: 2]"
                      }
           --> Network test will have addition flags like ipv6Server. Also the
               default num_vm is overwritten to 2. test_ids that contains
               "Network" has additional two params
           "stress": {"--num_vm=<int>": "Number of vm [default: 12]"}
           --> Stress test will have default num_vm is 12. The order of test is
               from top to bottom. If Stress test above Network test, default
               num_vm is 2. If Stress test after Network test, default num_vm
               is 12

        If there is the conflict in the parameters, the last parameter will
        take the affect. Example:
        "--num_vm=<int>": "number of VMs [default: 2]"
        "Network": {"--num_vm=<int>": "number of VMs [default: 4]"
        --> flag num_vm will have default value 4

        @param jsonfile: json file that has the structure as describe above
        @return: True for successful parsing json file, False otherwise
        """
        with open(jsonfile) as f:
            # Using object_pairs_hook=OrderedDict to keep the ordering of json
            json_data = json.load(f, object_pairs_hook=OrderedDict)
        for k, val in json_data.items():
            if not self._add_argument(k, val):
                self._json_param[k] = val

    def _get_default(self, default_str, metadata, argument):
        """
        Function helper for _add_argument to get the default value
        """
        retval = default_str.split(":")[1].strip(']').strip()
        if retval:
            if metadata == 'int':
                retval = int(retval)
        #if argument[-1] == 's':
        #    return [retval, ]
        return retval

    def _get_choices(self, choice_str, metadata='str'):
        """
        Function helper for _add_argument to get the choice options
        """
        choice_str = choice_str.split(':')[1].strip(']').strip()
        if metadata == 'int':
            choices = [int(item.strip()) for item in choice_str.split()]
        else:
            choices = [item.strip() for item in choice_str.split()]
        return choices

    def _add_argument(self, argument, helpValue):
        """
        Add param and help string. Argument should start with "--"
        Function will return False if argument doesn't start with "--"
        Argument contain this pattern "=<>" require input from command line.
        For example:
        --Driver=<str>: flag Driver need input string --Driver e1000e
        --skip_cleanup: flag skip_cleanup does not need input and default is
                       False
                       if --skip_cleanup added in the command line, flag
                       skip_cleanup is True
        --runtime=<int>: flag runtime need input integer --runtime 360

        If argument string end with 's', it can take more than one value
        separated by spaces
        For example:
        --Drivers=<str>: flag Drivers need input string and take more than
                         one value --Drivers e1000e bxn

        Help value is the string helper for argument. In addition, helpValue
        can specify the default value for flag, the flag is required or not
        and choice option:
        For example:
        "--VMDK=<str>": "Test vmdk [default: RHEL5_32]"
                        --> VMDK flag has default value is "RHEL5-32". This
                        flag is not required because it has default value
        "--MTU=<int>": "MTU setting for test [default: 1500] [choice: 1500,
                        6000, 9000]"
                        --> MTU flag has default value integer 1500, it will
                        raise error if command line input something that not
                        in choice list. This flag is not required because it
                        has default value
        "--Driver=<str>: "Test Driver on host [required]"
                         --> Driver flag need input for driver and this flag
                          is required
        When help string contain both [default: <value>] and [required] flag,
        this function will throw ValueError because these flags are exclusive

        Flags without input is always boolean type. It's not a required flag
        and default value is always False. If add the flag, value is True

        @param argument: flag argument
        @param helpValue: help information the flag that include default and
                          choice
        @return: True successful parse the argument. False if the argument
                 does not start with '--'
        @raise ValueError: if fail to parse
        """
        special_pattern = '\[.*?\]'
        arg_val = {}
        required_str = '[required]'

        if not argument.startswith('--'):
            return False

        # parse special pattern
        specials = re.findall(special_pattern, helpValue)
        required_flag = required_str in specials
        # required string is removed in help message
        helpValue = helpValue.replace(required_str, '')

        # parse metavar. Metavar is in argument that has format --Flag=<str>
        # valid metavar are int and str
        if '=' in argument:
            argument, metavar = argument.split('=')
            metavar = metavar.strip('<>')
            if metavar == 'int' or metavar == 'str':
                arg_val['metavar'] = metavar
                if metavar == 'int':
                    arg_val['type'] = int
            else:
                raise ValueError('Invalid meta variable ({}). Input: {}={}\n'
                                 'Metavar should be <int>, <str> or empty'
                                 .format(metavar, argument, metavar))
            for item in specials:
                if 'default:' in item:
                    # default tag still say in help message
                    arg_val['default'] = self._get_default(item, metavar, argument)
                elif 'choice:' in item:
                    # choice and require tag will be removed in help message
                    arg_val['choices'] = self._get_choices(item, metavar)
                    arg_val['metavar'] = None
                    helpValue = helpValue.replace(item, '')
        else:
            arg_val['action'] = 'store_true'
        # Add help
        arg_val['help'] = helpValue
        # If argument end with 's', this argument can take more than 1 value
        if argument[-1] == 's':
            arg_val['nargs'] = '+'
        # Add into required group or optional group
        if required_flag:
            if arg_val.get('default'):
                raise ValueError('Required parameter do not have default value. Please check {}.\n'
                                 'Help string should not contain both [default] and [required]'
                                 .format(argument))
            arg_val['required'] = True
            self._req_group.add_argument(argument, **arg_val)
        else:
            # check for default value should be in choices
            default_val = arg_val.get('default')
            choice_vals = arg_val.get('choices')
            if default_val and choice_vals and default_val not in choice_vals:
                raise ValueError('Default is %s but choices are: %s'
                                 % (default_val, ', '.join(choice_vals)))
            self._ops_group.add_argument(argument, **arg_val)
        return True

    def parse_args(self):
        # get the test_ids arguments to get the test cases
        test_id = ''
        try:
            for idx, arg in enumerate(sys.argv):
                if 'test_id' in arg:
                    test_id = sys.argv[idx + 1]
                    break
        except IndexError:
            test_id = ''

        # Add additional arguments depend on test cases
        for k, arg_dicts in self._json_param.items():
            if k.lower() in test_id.lower():
                for arg, val in arg_dicts.items():
                    if not self._add_argument(arg, val):
                        # Fail to add argument, it must be Example
                        self._parser.epilog = '{}: {}'.format(arg, val)

        # parse arguments again with test cases added
        args = self._parser.parse_args()

        # converting namespace to dict
        self._args = vars(args)


# ----------------------------------------------------------------------------
#                TEST BLOCK
# ----------------------------------------------------------------------------
if __name__ == '__main__':
    bindir = os.path.dirname(os.path.abspath(__file__))
    myparser = ArgParser('%s/TestLauncherParam.json' % bindir)
    try:
        myparser.parse_args()
    except ValueError as e:
        print('Command line error\n{}.'.format(e))
        sys.exit(2)
    for k, val in myparser.args.items():
        print('key=%s, value=%s' % (k, val))
