#!/bin/env python
##############################################################################
# Copyright 2015-2017 VMware, Inc.  All rights reserved. VMware Confidential
##############################################################################
import sys
import traceback
import os
import re
import time
import socket
import signal
from operator import itemgetter
import subprocess
import argparse
import json
import threading
import logging.config
from lib.Util import get_esx_boot_type, genWaitTime, ping, PSOD_MSG, SETUP_MSG
from lib.ReVal import ReVal
from lib.ESXHandler import ESXHandler
from remotes.STAFRemote import STAFRemote

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
console = logging.StreamHandler(sys.stdout)
log.addHandler(console)
DEFAULT_CFG_FILE = 'TestGroups.json'


class TestGroupError(Exception):
    pass


def handle_signal(signum, frame):
    """
    Handle signal ternimate from Jenkins
    Generating SystemExit from python so that TestGroup can handle cleanup properly
    """
    print('Signal handler called with signal %s' % signum)
    raise SystemExit()


class TestResults(object):
    """
    class to report test results
    """
    # If TestLauncher return one of these return code, Test Group will consider
    # it as skip tests
    SKIP_LIST = (ReVal.NOT_ENOUGH_TEST_NIC,
                 ReVal.ERROR_INFO_IN_XML,
                 ReVal.DRIVER_NOT_SUPPORTED,
                 ReVal.MANUAL_TEST,
                 ReVal.NOT_ENOUGH_TEST_HOST,
                 ReVal.MISSING_VC,
                 ReVal.NOT_RUN,
                )

    RESULT_COLORS = {
                ReVal.PASSED : '#00ff00',   # light green
                ReVal.FAILED : 'red',
                ReVal.INVALID_CMD : 'red',
                ReVal.PLATFORM_NOT_SUPPORTED : 'gray',
                ReVal.VM_OS_NOT_SUPPORTED : 'gray',
                ReVal.VNIC_TYPE_NOT_SUPPORTED : 'gray',
                ReVal.VMNIC_DRIVER_NOT_PRESENT : 'red',
                ReVal.SHORT_OF_MANDATORY_PARAMS : 'gray',
                ReVal.ERROR_INFO_IN_XML : 'gray',
                ReVal.NOT_ENOUGH_TEST_NIC : 'gray',
                ReVal.NOT_ENOUGH_TEST_HOST : 'gray',
                ReVal.NAS_CONFIG_FAILED : 'red',
                ReVal.DRIVER_NOT_SUPPORTED : 'gray',
                ReVal.ENV_FAILED : 'red',
                ReVal.SCRIPT_ERROR : 'red',
                ReVal.ZDUMP_FOUND : 'red',
                ReVal.UPLINK_ERROR : 'red',
                ReVal.DATA_CORRUPTION : 'red',
                ReVal.VOMA_ERROR : 'red',
                ReVal.LOG_SPEW : 'red',
                ReVal.ESX_PSOD : 'red',
                ReVal.CLEAN_ENV_FAILED : 'red',
                ReVal.SETUP_ENV_FAILED : 'red',
                ReVal.NOT_RUN : 'white',
                ReVal.MANUAL_TEST : 'gray',
                ReVal.MISSING_VC : 'gray',
                ReVal.STAF_FAILED : 'red',
                ReVal.STAF_HOST_CMD_FAILED : 'red',
                ReVal.STAF_UNMARSHALL_FAILED : 'red',
                ReVal.ABORTED : 'pink',
                ReVal.PASSED_NOT_ENOUGH_TIME : 'yellow'
            }

    def __init__(self, tcs):
        """
        main initialization for TestResult
        @param tcs: a list of test case, the order of test case list must keep
                    the same for entire test
                    Each entry in test case list is a dictionary with keys:
                    'TestID', 'setup', 'cleanup', 'SetupType'
                    The list is sort in order of 'SetupType'
        """
        self.tcs = [{'test_id': tc['TestID'], 'result': ReVal.NOT_RUN,
                     'start': 'N/A', 'end': 'N/A', 'elap': 'N/A', 'logpath': None}
                     for tc in tcs]

    def __getitem__(self, index):
        """
        get test case
        @param index:
        @return: test case dictionary
        """
        return self.tcs[index]

    def _print_html(self, JenkinsBuildNo, totaltests, fail_cases, pass_cases, skip_cases, missed_cases):
        """
        This function prints html report
        @param jenkinsBuildNo: jenkins build number or a unique number to print a report
        """
        try:
            import HTML
        except ImportError:
            log.warn('Failed to import HTML package, report will not be generated')
            return

        logDirPrefix = 'http://' + socket.gethostname()
        t = HTML.Table(header_row=['Test Name', 'Status', 'Start time', 'End time', 'Duration', 'Log Folder'])
        for test_res in self.tcs:
            logpath = test_res['logpath']
            if logpath:
                # get the entire logs folder, not just main log
                logpath = '<a href=' + logDirPrefix + logpath + '>logs</a>'
            # Define the table cell color based on the test result
            color = TestResults.RESULT_COLORS[test_res['result']]
            colored_result = HTML.TableCell(ReVal.toString[test_res['result']], bgcolor=color)
            t.rows.append([test_res['test_id'], colored_result,
                    test_res['start'], test_res['end'], test_res['elap'], logpath])

        # Add a summary to the end of the table
        total = 'Total tests: %s' %totaltests
        failures = 'Failed: %s' %fail_cases
        passes = 'Passed: %s' %pass_cases
        skips = 'Skipped: %s' %skip_cases
        noruns = 'Not Run: %s' %missed_cases
        logpath = ' '
        t.rows.append([total, passes, failures, skips, noruns, logpath])
        htmlcode = str(t)
        reportDir = 'TestReports'
        if not os.path.exists(reportDir):
            os.makedirs(reportDir)
        # report goes to the same directory to where it is launched from
        with open('%s/report-%s.html' % (reportDir, JenkinsBuildNo), 'w') as hr:
            hr.write(htmlcode)

    def report_test_group_results(self, jenkins_bld, print_table=True):
        """
        Print a short report table, also print html to link to test result
        @param jenkins_bld: jenkins build id
        @param print_table: True/False flag for printing table report
        """
        totaltests = len(self.tcs)
        fail_cases, pass_cases, skip_cases, missed_cases = 0, 0, 0, 0
        lines = []
        width = 80
        lines.append('Test Results'.center(width, '='))
        for item in self.tcs:
            try:
                lines.append('%s --- %s' % (item['test_id'], ReVal.toString[item['result']]))
                if item['result'] == ReVal.PASSED:
                    pass_cases += 1
                elif item['result'] != ReVal.NOT_RUN:
                    if item['result'] in TestResults.SKIP_LIST:
                        skip_cases += 1
                    else:
                        fail_cases += 1
                else:
                    missed_cases += 1
            except KeyError as e:
                # this happend when Test is manually get kill
                # will mark case as failure
                log.warning('Test %s received unexpected rc %s' % (item['test_id'], e))
                item['result'] = ReVal.FAILED
                lines.append('%s --- %s' % (item['test_id'], ReVal.toString[item['result']]))
                fail_cases += 1

        lines.append(''.center(width, '='))
        lines.append('Total Tests: %s, Passed: %s, Failed: %s, Skipped: %s Not Run: %s'
                     % (totaltests, pass_cases, fail_cases, skip_cases, missed_cases))
        if print_table:
            log.info('\n%s' % '\n'.join(lines))
        self._print_html(jenkins_bld, totaltests, fail_cases, pass_cases, skip_cases, missed_cases)
        if fail_cases or not pass_cases:
            return ReVal.FAILED
        else:
            return ReVal.PASSED


class TestGroups:
    """
    Read test group info from command line and parses given test group xml or
    default xml file and create a structure with applicable tests
    and properties
    """
    # extra waiting time for in seconds for a running TestLauncher before
    # test group kills it
    TIMEOUT_EXTRA = 120 * 60

    def __init__(self, args, binpath):
        """
        Initialize module for TestGroup
        """
        self.jenkins_bld = args.jenkinsBuild
        self.hosts = args.testhost
        self.drivers = args.drivers
        self.cfg_file = args.group_cfg_file
        self.testlauncher_path = os.path.join(binpath, 'TestLauncher.py')
        if args.testList:
            test_ids = [test['testcase'] for test in json.loads(args.testList)]
        else:
            test_ids = None
        self.testCaseList = self._sort_tests(self._get_test_from_json(args.group_cfg_file,
                                                                      args.testgroup, test_ids))
        tcs_str = ['{0:56} SetupType: {1:24} setup: {2:8} cleanup: {3}'.
                   format(tc['TestID'], tc['SetupType'], str(tc['setup']), tc['cleanup'])
                   for tc in self.testCaseList]
        log.info('Test Case list:\n%s\n\n' % '\n'.join(tcs_str))
        self.test_results = TestResults(self.testCaseList)

    def _get_test_from_json(self, json_file, group=None, test_ids=None):
        """
        Return a list of test. Each test is the dictionary describe test
        name, test setup type, duration of the test, valid drivers, etc
        @param json_file: json file name. This file is generated from
                          matrix excel file
        @param group: name of test group to run. If this parameter is provided,
                      test_ids should be empty
        @param test_ids: a list of test Ids, if this parameter is provided,
                         group should be empty. This parameter for selected
                         test
        @return: a list of test. Each test is a dictionary
        """
        if group and test_ids:
            raise TestGroupError('Not support both parameters test group and test select.')
        # Json file is expected to have 2 dictionaries, the first one is test group
        # and the second one is test cases.
        # Test group dictionary with key are test group name and values are the list
        # of test name
        # Test case dictionary with key are test id and values are another dictionary
        # describe the test
        with open(json_file) as f:
            groups, test_cases = json.load(f)
        if test_ids is None:
            try:
                test_ids = groups[group]
            except KeyError:
                raise TestGroupError('Invalid group name: %s. Valid groups are %s'
                                     % (group, ', '.join(groups)))
        test_ids_lower = [id.lower() for id in test_ids]
        retval = [val for id, val in test_cases.items() if id in test_ids_lower]
        return retval

    def _sort_tests(self, test_cases):
        """
        Divide lists based on setupType and merge after necessary setup and
        cleanup flags were set
        @param test_cases:
        """
        sorted_test_cases = sorted(test_cases, key=itemgetter('SetupType'))
        prev_setup = None
        for idx, test in enumerate(sorted_test_cases):
            curr_setup = test['SetupType']
            try:
                next_setup = sorted_test_cases[idx+1]['SetupType']
            except IndexError:
                next_setup = None
            test['setup'] = curr_setup != prev_setup
            test['cleanup'] = curr_setup != next_setup
            prev_setup = curr_setup
        return sorted_test_cases

    def _is_mult_hosts(self, test_id, test_setup):
        """
        Return True/False for test require multiple hosts
        @param test_id: test case ID
        @param test_setup: test setup type
        @return: True for test requires multiple hosts, False otherwise
        """
        # Networking tests with multiple hosts, 'TwoHosts' will be in
        # the test setup type
        if 'twohost' in test_setup.lower():
            return True
        else:
            return False

    def _gen_cmd(self, args, tc, setup, cleanup, multi_hosts):
        """
        Internal method to create test command
        @param args: test argumenets
        @param tc: test case dictionary
        @param setup: True/False
        @param cleanup: True/False
        @param multi_hosts: True/False
        @return: test command string
        """
        cmds = [self.testlauncher_path]
        addOps = ['--test_ids', tc['TestID']]
        if not setup:
            addOps.append('--skip_setup')
        if not cleanup:
            addOps.append('--skip_cleanup')
        # adding back driver and testgroup parameters
        args = ['--drivers'] + self.drivers + args + addOps
        if self.cfg_file != DEFAULT_CFG_FILE:
            args += ['--group_cfg_file', self.cfg_file]
        if multi_hosts:
            return cmds + self.hosts + args
        else:
            return cmds + self.hosts[0:1] + args

    def _is_special_test(self, testId, host):
        """
        Check to see if setup needs to be created for special test cases
        Special test case that required test to be rebooted and the host
        is pxe boot, in this case, all the setup before reboot will be
        removed
        @param testId: test name
        @param host: host name string
        @return: True/False
        """
        testId = testId.lower()
        splTests = ('dump', 'ipxe', 'reboot')
        for key in splTests:
            if key in testId:
                if get_esx_boot_type(host) == 'pxe':
                    return True
        else:
            return False

    def collect_vm_support(self, log_path):
        """
        Collect vm-support for all tested hosts
        @param log_path: location on launcher to store the vm-support tar file
        """
        def _run_vm_support(host, log_path):
            esx = ESXHandler(host)
            outfile = esx.vm_support()
            if esx.remote.CopyFromHost(outfile, log_path):
                esx.rm(outfile)
            else:
                log.warning('Copy %s to launcher failed' % outfile)

        log.info('Collecting vm_support for %s' % (', '.join(self.hosts)))
        threads = [threading.Thread(target=_run_vm_support, args=(host, log_path), daemon=True)
                   for host in self.hosts]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

    def handle_psod(self, host_name, log_path):
        # wait until host is up, then continue
        wait_arry = genWaitTime(1800, 1200)
        for duration in wait_arry:
            if ping(host_name):
                remote = STAFRemote()
                # wait again for STAF up
                for staf_duration in genWaitTime(120):
                    if remote.StafPing(host_name):
                        self.collect_vm_support(log_path)
                        break
                    else:
                        log.info('Waiting for STAF up...')
                        time.sleep(staf_duration)
                else:
                    raise TestGroupError('STAF is not up on %s' % host_name)
                break
            else:
                log.info('%s is PSOD, waiting for server rebooted...' % host_name)
                time.sleep(duration)
        else:
            raise TestGroupError('%s is PSOD and not able to recover' % self.hosts[0])

    def _run_test_helper(self, report_entry, cmd, timeout, setup):
        """
        Helper to run a test and report the result
        @param report_entry: a dictionary entry to report the test status
        @param cmd: actual command list to run
        @param timeout: in seconds
        @param setup: test needs to do setup
        @return: True for test successfully setup, False otherwise
        @raise: TestGroupError when host is PSOD or issue with the command line
        """
        terminated, psod = False, ''
        output = []
        psod_pattern = PSOD_MSG.format(server='(\S+)')
        starttime = time.time()
        report_entry['start'] = time.strftime('%Y-%m-%dT%H:%M:%S', time.localtime())
        log.info('Test Command: %s' % ' '.join(cmd))
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        for line in iter(process.stdout.readline, ''):
            line = line.decode('utf-8')
            output.append(line)
            sys.stdout.write(line)
            if line == '' and process.poll() is not None:
                break
            # Monitor PSOD string, if there is, kill the test immediately
            # However some test creates PSOD using CrashPanic function
            psod_matched = re.search(psod_pattern, line)
            if psod_matched and 'CrashPanic' not in output[-2]:
                # a host is PSOD, wait a minute so that serial logs
                # can collect all traces
                time.sleep(60)
                os.kill(process.pid, signal.SIGKILL)
                psod = psod_matched.group(1)
            if timeout:
                if time.time() - starttime > timeout and not terminated:
                    log.warning('Killing test %s because it has been running for too long'
                                % report_entry['test_id'])
                    terminated = True
                    os.kill(process.pid, signal.SIGINT)
        rc = process.returncode if not psod else ReVal.ESX_PSOD
        report_entry['end'] = time.strftime('%Y-%m-%dT%H:%M:%S', time.localtime())
        report_entry['elap'] = time.strftime('%H:%M:%S', time.gmtime(time.time() - starttime))
        report_entry['result'] = rc

        # concat all the output lines, search for special pattern in the output
        # to get the location of the log folder and make sure the test setup actually
        # pass or fail. We are using string search for now, not sure if there
        # is the better way to handle this
        output = ''.join(output)
        match = re.search('.*Setting logs folder: (.*)', output)
        if match:
            self.test_results[index]['logpath'] = match.group(1)
        elif not rc:
            # Test passed but not able to find log folder
            log.warning('Unable to locate test logs folder')

        # handle error case, should output error message
        if rc:
            err_str = ''.join([line.decode('utf-8') for line in process.stderr.readlines()])
            sys.stdout.write(err_str)
            if rc == ReVal.INVALID_CMD:
                # invalid commands, test group should bail immediately
                raise TestGroupError('Invalid TestLauncher.py commands.')

            # When test fails because of test name does not exits or driver does
            # does not support or test missing VC, the logic to skip setup True/False
            # should remain the same as the previous test
            # For example: we run 3 test: test1, test2, test3 in the same setup
            # test1 successful setup, test2 fails because it's not defined,
            # test3 should be able to run without setup
            if rc in TestResults.SKIP_LIST:
                # if setup is True, we should return False (previous setup failed)
                # if setup is False, we should return True (previous setup passed)
                return not setup

        setup_success = SETUP_MSG in output
        if setup_success:
            if rc == ReVal.ESX_PSOD or self._is_special_test(tc['TestID'], self.hosts[0]):
                setup_success = False
        if rc == ReVal.ESX_PSOD:
            # TODO: only handle PSOD on the first host
            # need to handle on other hosts
            self.handle_psod(psod, match.group(1))
        return setup_success

    def run_test(self, index, tc, args, previous_setup=False):
        """
        Main code to run the TestLauncher
        @param index: Since test is run as a list, it had index number
        @param tc: test case dictionary.
        @param args: arguments to run the test
        @param previous_setup: flag for the previous setup pass or fail
        @return: test setup pass or fail (True/false)
        @raise: TestGroupsError
        """
        def _is_reboot(testId):
            testId = testId.lower()
            splTests = ('dump', 'ipxe', 'reboot', 'createvmfs')
            for key in splTests:
                if key in testId:
                    return True
            else:
                return False

        setup = tc['setup']
        validDrivers = tc.get('validDrivers')
        """
        if _is_reboot(tc['TestID']):
            log.info('%s need to reboot, skip for now...' % tc['TestID'])
            # if this the first test of the group, we don't want to return
            # the previous setup value.
            # To know if this is the first test, setup should be True,
            # in this case, return false
            if setup:
                return False
            else:
                return previous_setup
        """
        mult_host = self._is_mult_hosts(tc['TestID'], tc['SetupType'])

        # If previous test's setup failed and the current test is assuming
        # previous test's setup was successful and skipping it, that's not
        # right, below code will fix it
        if not setup and not previous_setup:
            setup = True

        # Only check for the first driver for now...
        if validDrivers:
            if self.drivers[0].replace(ESXHandler.UNSTABLE, '') not in validDrivers:
                log.warning('Driver %s is not suitable for this test %s'
                            % (self.drivers[0], tc['TestID']))
                self.test_results[index]['result'] = ReVal.DRIVER_NOT_SUPPORTED
                # If this is a skip-setup then previous setup value should be returned
                if not setup:
                    return previous_setup
                else:
                    return False

        # check if test multiple hosts or not
        if mult_host:
            if len(self.hosts) == 1:
                log.warn('Test %s requires multiple hosts. Current TestGroups has only one host'
                         % tc['TestID'])
                self.test_results[index]['result'] = ReVal.NOT_ENOUGH_TEST_HOST
                if not setup:
                    return previous_setup
                else:
                    return False

        # figure timeout
        try:
            # get the runtime from the command output.
            timeout_cmd = int(args[args.index('--runtime') + 1]) + TestGroups.TIMEOUT_EXTRA
        except ValueError:
            timeout_cmd = 0
        try:
            # Using the runtime from excel file
            # This value is in minute unit. Need to convert into seconds
            timeout_xml = int(tc['Test execution duration'].split()[0]) * 60 + TestGroups.TIMEOUT_EXTRA
        except KeyError:
            timeout_xml = 0

        timeout = max(timeout_cmd, timeout_xml)

        # TODO: remove
        # This is the workaround until we fix the excel sheet
        if 'manual' in tc['SetupType'].lower():
            log.warn('%s is the manual test. Skip it...' % tc['TestID'])
            self.test_results[index]['result'] = ReVal.toString[ReVal.MANUAL_TEST]
            if not setup:
                return previous_setup
            else:
                return False

        command = self._gen_cmd(args, tc, setup, tc['cleanup'], True)
        previous_setup = self._run_test_helper(self.test_results[index], command, timeout, setup)
        return previous_setup


#------------------------------------------------------------------------------
# Main processing from here
#------------------------------------------------------------------------------
if __name__ == '__main__':
    # Simple argument parser for main
    parser = argparse.ArgumentParser(description='TestGroups argument parser')
    parser.add_argument('testhost', help='Primary ESX host name or ip address', nargs='+')
    parser.add_argument('--drivers', help='Test drivers', nargs='+')
    parser.add_argument('--testgroup', help='Test Group name')
    parser.add_argument('--jenkinsBuild', default=0,
                        help='Jenkins build number or a unique string to generate report')
    parser.add_argument('--group_cfg_file', default=DEFAULT_CFG_FILE,
                        help='Test grouping configuration file (default: %s)' % DEFAULT_CFG_FILE)
    parser.add_argument('--testList', help='If Test Group above is None, please provide '
                                           'dict of list of test cases')
    args, unknownArgs = parser.parse_known_args()

    # trapping kill signals here
    signal.signal(signal.SIGTERM, handle_signal)

    start_time = time.time()
    tg = None
    try:
        binpath = os.path.dirname(os.path.abspath(__file__))
        tg = TestGroups(args, binpath)
        # generate test cases and run each test case list
        previous_setup = False
        for index, tc in enumerate(tg.testCaseList):
            tg.test_results.report_test_group_results(tg.jenkins_bld, print_table=False)
            # add extra line just for separation of tests
            print('\n')
            previous_setup = tg.run_test(index, tc, unknownArgs, previous_setup)
    except TestGroupError as e:
        log.error('There is an error, check logs, exception message: %s' % e)
    except KeyboardInterrupt:
        log.info('Test halted')
    except Exception as e:
        traceback.print_exc(file=sys.stdout)
    finally:
        if tg:
            rc = tg.test_results.report_test_group_results(tg.jenkins_bld)
            total_runtime = time.time() - start_time
            log.info('Total test group runtime is: %s '
                     % time.strftime('%H:%M:%S', time.gmtime(total_runtime)))
            sys.exit(rc)
