#! /usr/bin/python

##############################################################################
# Copyright 2014-2017 VMware, Inc.  All rights reserved. VMware Confidential
##############################################################################
"""
This is the main module to Launch Test
Usage: TestLauncher <host> --Test_ID <test> [options]...

options:
--skip_setup:   TestLauncher will assume the NAS, networking and VMs are ready
                and the setup will be ignored
--skip_cleanup: Reserve the environment, TestLauncher will not clean up the
                environment, the test bed can be reserved for the next test
                which has the same testing environment as the former.
"""
from lib.HostProfileHandler import HostProfileHandler
from lib.NetworkingHandler import NetworkingHandler
from lib.VxlanHandler import VxlanHandler
from lib.StorageHandler import StorageHandler
from lib.HotplugHandler import HotplugHandler
from lib.FCHandler import FCHandler
from lib.vSANHandler import vSANHandler
from lib.DDVHandler import DDVHandler
from lib.FCoEHandler import FCoEHandler
from lib.CATFCoEHandler import CATFCoEHandler
from lib.iSCSIHandler import iSCSIHandler
from lib.iSERHandler import iSERHandler
from lib.NoVMNetworkHandler import NoVMNetworkHandler
from lib.NoVMStorageHandler import NoVMStorageHandler
from lib.GeneveHandler import GeneveHandler
from lib.SRIOVHandler import SRIOVHandler
from lib.IODMHandler import IODMHandler
from lib.USBHandler import USBHandler
from lib.DifDixHandler import DifDixHandler
from lib.RoCEHandler import RoCEHandler
from lib.AsyncDriverHandler import AsyncDriverHandler
from lib.TestHandler import TestHandler, SkipConfigure
from lib.ConsCheck import ConsCheck
from lib.HweLogging import setup_logging, create_log_folder, VmkLogHandler, StackHandler
from remotes.STAFRemote import STAFRemote
from lib.ReVal import ReVal
from lib.Util import BANNER_WIDTH, ping, TEST_CLEANUP, SETUP_MSG
from lib.HweExceptions import CustomExitError, LogicExit, EnvironmentError
from ArgParser import ArgParser
import sys
import os
import logging
import atexit
import time
import signal


log = logging.getLogger(__name__)


def printBanner():
    """
    This function is just to print the Test banner
    """
    log.info("*" * BANNER_WIDTH)
    log.info("Welcome to the HWE Test Automation Framework."
             .center(BANNER_WIDTH))
    log.info("Copyright 2014-2017 VMware, Inc.  All rights reserved."
             .center(BANNER_WIDTH))
    log.info("*" * BANNER_WIDTH)


# Adding a global dictionary Test case and Test result here
# We will have multiple test cases and the test result for each test case
TestCaseResults = {TEST_CLEANUP: ReVal.PASSED, }


# Handle signal terminate to kill test gracefully
def handlesignal(signum, frame):
    """
    Trapping signal terminate from Jenkins so that python can
    cleanup environment properly
    """
    raise KeyboardInterrupt()


#------------------------------------------------------------------------------
# Main processing from here
#------------------------------------------------------------------------------
if __name__ == "__main__":
    bindir = os.path.dirname(os.path.abspath(__file__))
    test_parser = ArgParser(os.path.join(bindir, 'TestLauncherParam.json'))
    test_parser.parse_args()
    currentTest = test_parser.args['test_ids'][0]
    datestamp = time.strftime('%d-%b-%Y')
    timestamp = time.strftime('%H_%M_%S')

    signal.signal(signal.SIGTERM, handlesignal)
    consCheck, myTest = None, None
    exit_status = ReVal.NOT_RUN
    atexit.register(STAFRemote.destroyHandles)

    try:
        logfolder = create_log_folder(test_parser.args, datestamp, timestamp)
        setup_logging(os.path.join(bindir, 'logging.json'), logfolder)
        StackHandler.PRINT_STACK = test_parser.args['print_stack']
        printBanner()
        log.info('Setting logs folder: %s' % logfolder)

        consCheck = ConsCheck(bindir, test_parser.args, logfolder)

        atexit.register(consCheck.createSummaryReport, TestCaseResults)

        # SetHostName should be called after consCheck
        # before it calling STAF remote. STAF might not be
        # initialized before ConsCheck
        VmkLogHandler.ENABLED = True

        # Get parameter from command line and setup consCheck
        consCheck.StartSerialLogging()
        consCheck.CheckLogDisk(test_parser.args['log_location'])
        consCheck.SyncTime()

        # create shadow vmklog copy
        consCheck.StartKernelLog()

        moduleToTestHandle = {'ddv': DDVHandler,
                              'network': NetworkingHandler,
                              'storage': StorageHandler,
                              'fc':FCHandler,
                              'iscsi': iSCSIHandler,
                              'iser': iSERHandler,
                              'FCoE': FCoEHandler,
                              'cat_fcoe': CATFCoEHandler,
                              'RoCE': RoCEHandler,
                              'novm': NoVMNetworkHandler,
                              'novm_storage': NoVMStorageHandler,
                              'iodm': IODMHandler,
                              'difdix': DifDixHandler,
                              'vxlan': VxlanHandler,
                              'vsan': vSANHandler,
                              'asyncdriver': AsyncDriverHandler,
                              'geneve': GeneveHandler,
                              'sriov':SRIOVHandler,
                              'fcoeplugin': HostProfileHandler,
                              'usb': USBHandler,
			      'hotplug': HotplugHandler,
                              }
        chosen_handler = moduleToTestHandle.get(consCheck.ioType, TestHandler)
        myTest = chosen_handler(consCheck)
        try:
            myTest.Setup()
            if not myTest.ConfigTest():
                raise EnvironmentError('Failed to Configure Test')
        except SkipConfigure:
            pass

        # If you are here means Setup or config was done successfully
        log.info(SETUP_MSG)

        padding = 15
        test_str_len = padding + len(consCheck.testId)
        log.info("*" * test_str_len)
        log.info("Starting Test %s" % consCheck.testId)
        log.info("*" * test_str_len)

        testFunc = getattr(myTest, consCheck.testfunc)
        consCheck.dtaas_start()
        exit_status = testFunc()
        consCheck.dtaas_stop(exit_status)
        TestCaseResults[TEST_CLEANUP] = myTest.Cleanup(exit_status)
    except CustomExitError as e:
        log.error(e)
        exit_status = e.exit_status
        if isinstance(e, EnvironmentError):
            myTest.Cleanup(exit_status)
    except LogicExit as e:
        log.error(e)
        exit_status = ReVal.FAILED
        myTest.Cleanup(exit_status)
    except Exception as e:
        exit_status = ReVal.SCRIPT_ERROR
        log.error('Test ran into exception', exc_info=True)
        if myTest:
            if not ping(myTest.esxhost.hostname):
                log.error('Failed to ping %s. Check server for PSOD'
                          % myTest.esxhost.hostname)
                exit_status = ReVal.ESX_PSOD
            else:
                myTest.Cleanup(exit_status)
    except KeyboardInterrupt:
        log.info('Received keyboard interrupt during test')
        exit_status = ReVal.ABORTED
    finally:
        TestCaseResults[currentTest] = exit_status
        # If test pass but cleanup fails, TestLauncher will return fail
        if exit_status == ReVal.PASSED and TestCaseResults[TEST_CLEANUP] != ReVal.PASSED:
            exit_status = ReVal.FAILED
        if consCheck:
            consCheck.dtaas_stop(exit_status)
            if consCheck.serial_logging:
                consCheck.serial_logging.stopSerialCapture()
            if exit_status != ReVal.ESX_PSOD:
                consCheck.StopKernelLog()
            consCheck.PostToHPQC(TestCaseResults)
        sys.exit(exit_status)
